// Copyright (c) 2017 Douglas Miller <durgadas311@gmail.com>

// Emulates Jonathon Nock's ATmega328 "system support module",
// may also work for v2 module based on STM32.

import java.io.*;
import java.util.Date;
import java.text.SimpleDateFormat;

// General command/response format:
//
//	%L
//	%L<adr>#
//	%L<adr><string>#
//
//	"L" represents the command letter.
//	<adr> is a 4-digit hex string
//	<str> is an arbitrary (?) length string of ASCII characters
//	Command letter determines format of command
//
// Supported commands:
//	Command		Function		Response
//	--------------	----------------------	--------------
//	%R		Reset the system   	none
//	%S		Get CPU clock speed	%S[248]#
//	%C		toggle CPU clock speed	none
//	%T		Print time/date		%T<time>#
//	%Z<adr>#	Get time/date		%E<adr><time>#
//	%U<adr><time>#	Set time/date		%E<adr><time>#
//
// Generic response commands:
//	%E<adr><str>#	Put <str> into memory at <adr>
//	%T<str>#	Echo <str> to terminal output
//	%Q<str>#	Send <str> to terminal input (as a command)
//
//	<time> format is "dd/MM/yyyy HH:mm:ss EEE"
//
public class ATmega328 implements Runnable {
	VirtualUART uart;
	Z80mcCommander mc;
	enum States {
		IDLE,
		COMMAND,
		PARAM,
	};

	private States state;
	private String str;
	private int prm;
	private int pc;
	private int cmd;
	public static SimpleDateFormat timestamp =
		new java.text.SimpleDateFormat("dd/MM/yyyy HH:mm:ss EEE");

	public ATmega328(VirtualUART uart, Z80mcCommander mc) {
		this.uart = uart;
		this.mc = mc;
		state = States.IDLE;
		Thread t = new Thread(this);
		t.start();
	}

	private void putStr(String rsp) {
		for (byte b : rsp.getBytes()) {
			while (!uart.ready()) {
				if (uart.available() > 0) {
					int c = uart.take(); // keep things moving... sort of
				}
			}
			uart.put(b, false);
		}
	}

	private void systemReset() {
		mc.sendCommand("reset");
	}

	private void doRTC() {
		String tod = timestamp.format(new Date());
		tod = String.format("%%E%04x%s#", prm, tod);
		putStr(tod);
	}

	private void echoRTC() {
		String tod = timestamp.format(new Date());
		tod = String.format("%%T%s#", tod);
		putStr(tod);
	}

	private void processParam(int c) {
		if (c == '#') {
			processCmd();	// uses 'cmd' and 'prm'...
			state = States.IDLE;
			return;
		}
		if (pc < 4) {
			prm <<= 4;
			prm |= Character.digit((char)c, 16);
			++pc;
		} else {
			str += (char)c;
		}
	}

	private void initParam() {
		prm = 0;
		pc = 0;
		str = "";
		state = States.PARAM;
	}

	private void processCmd() {
		switch (cmd) {
		case 'U':	// set RTC time - ignored, mostly
		case 'u':	// (except it also sends back TOD)
System.err.format("ATmega328: set time %04x \"%s\"\n", prm, str);
			doRTC();
			break;
		case 'Z':	// get RTC time
		case 'z':
			doRTC();
			break;
		}
	}

	private void processCmd(int c) {
		cmd = c;
		state = States.IDLE;
		switch (c) {
		case 'C':	// toggle CPU clock speed
		case 'c':
			// no response...
			break;
		case 'R':	// reset the system...
		case 'r':
			// no response...
			systemReset();
			break;
		case 'S':	// get CPU clock speed
		case 's':
			putStr("%S4#");
			break;
		case 'T':	// echo time to terminal
		case 't':
			echoRTC();
			break;
		case 'U':	// set RTC time
		case 'u':
			initParam();
			break;
		case 'Z':	// get RTC time
		case 'z':
			initParam();
			break;
		default:
System.err.format("ATmega328: unknown command %02x\n", c);
			break;
		}
	}

	private void processChar(int c) {
		if (c == '%') {
			state = States.COMMAND;
		} else if (state == States.COMMAND) {
			processCmd(c);
		} else if (state == States.PARAM) {
			processParam(c);
		}
	}

	// This thread reads program stdout and sends to UART
	public void run() {
		while (true) {
			try {
				// This probably needs to be throttled...
				int c = uart.take();
				processChar(c);
			} catch (Exception ee) {
				ee.printStackTrace();
				uart.detach(); // repetative?
				break;
			}
		}
	}
}
